#
# Copyright 2018 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#

from __future__ import division
from __future__ import print_function

from math import atan, radians as rad
from pxr import Gf, Tf

from .qt import QtCore
from .common import DEBUG_CLIPPING

# FreeCamera inherits from QObject only so that it can send signals...
# which is really a pretty nice, easy to use notification system.
class FreeCamera(QtCore.QObject):

    # Allows FreeCamera owner to act when the camera's relationship to
    # its viewed content changes.  For instance, to compute the value
    # to supply for setClosestVisibleDistFromPoint()
    signalFrustumChanged = QtCore.Signal()
    signalFrustumSettingsChanged = QtCore.Signal()

    defaultNear = 1
    defaultFar = 2000000
    # When some of the scene is behind the camera, this is the largest fraction
    # of the scene still-in-front-of-the-camera that we are willing to clip
    # while computing a non-zero near-clipping plane
    maxAllowableSceneClipFactor = .01
    # Experimentally on Nvidia M6000, if Far/Near is greater than this,
    # then geometry in the back half of the volume will disappear
    maxSafeZResolution = 1e6
    # Experimentally on Nvidia M6000, if Far/Near is greater than this,
    # then we will often see Z-fighting artifacts even for geometry that
    # is close to camera, when rendering for picking
    maxGoodZResolution = 5e4

    def __init__(self, isZUp, fov=60.0, aspectRatio=1.0, overrideNear=None,
                 overrideFar=None):
        """FreeCamera can be either a Z up or Y up camera, based on 'zUp'"""
        super(FreeCamera, self).__init__()

        self._camera = Gf.Camera()
        self._camera.SetPerspectiveFromAspectRatioAndFieldOfView(
            aspectRatio, fov, Gf.Camera.FOVVertical)
        self._overrideNear = overrideNear
        self._overrideFar = overrideFar
        self.resetClippingPlanes()

        self._isZUp = isZUp

        self._cameraTransformDirty = True
        self._rotTheta = 0
        self._rotPhi = 0
        self._rotPsi = 0
        self._center = Gf.Vec3d(0,0,0)
        self._dist = 100
        self._camera.focusDistance = self._dist
        self._closestVisibleDist = None
        self._lastFramedDist = None
        self._lastFramedClosestDist = None
        self._selSize = 10

        if isZUp:
            # This is also Gf.Camera.Y_UP_TO_Z_UP_MATRIX
            self._YZUpMatrix = Gf.Matrix4d().SetRotate(
                Gf.Rotation(Gf.Vec3d.XAxis(), -90))
            self._YZUpInvMatrix = self._YZUpMatrix.GetInverse()
        else:
            self._YZUpMatrix = Gf.Matrix4d(1.0)
            self._YZUpInvMatrix = Gf.Matrix4d(1.0)

    # Why a clone() method vs copy.deepcopy()ing the FreeCamera ?
    # 1) Several of the Gf classes are not python-picklable (requirement of
    #    deepcopy), nor is GfCamera.  Adding that infrastructure for this
    #    single client seems weighty.
    # 2) We could make FreeCamera itself be picklable... that solution would
    #    require twice as much code as clone().  If we wind up extracting
    #    FreeCamera to be a more general building block, it may be worth it,
    #    and clone() would transition to __getstate__().
    def clone(self):
        clone = FreeCamera(self._isZUp)
        clone._camera = Gf.Camera(self._camera)
        # skipping stereo attrs for now

        clone._rotTheta = self._rotTheta
        clone._rotPhi = self._rotPhi
        clone._rotPsi = self._rotPsi
        clone._center = Gf.Vec3d(self._center)
        clone._dist = self._dist
        clone._closestVisibleDist = self._closestVisibleDist
        clone._lastFramedClosestDist = self._lastFramedClosestDist
        clone._lastFramedDist = self._lastFramedDist
        clone._selSize = self._selSize
        clone._overrideNear = self._overrideNear
        clone._overrideFar = self._overrideFar
        clone._YZUpMatrix = Gf.Matrix4d(self._YZUpMatrix)
        clone._YZUpInvMatrix = Gf.Matrix4d(self._YZUpInvMatrix)

        return clone


    def _pushToCameraTransform(self):
        """
        Updates the camera's transform matrix, that is, the matrix that brings
        the camera to the origin, with the camera view pointing down:
           +Y if this is a Zup camera, or
           -Z if this is a Yup camera .
        """
        if not self._cameraTransformDirty:
            return

        def RotMatrix(vec, angle):
            return Gf.Matrix4d(1.0).SetRotate(Gf.Rotation(vec, angle))

        # self._YZUpInvMatrix influences the behavior about how the
        # FreeCamera will tumble. It is the identity or a rotation about the
        # x-Axis.
        self._camera.transform = (
            Gf.Matrix4d().SetTranslate(Gf.Vec3d.ZAxis() * self.dist) *
            RotMatrix(Gf.Vec3d.ZAxis(), -self._rotPsi) *
            RotMatrix(Gf.Vec3d.XAxis(), -self._rotPhi) *
            RotMatrix(Gf.Vec3d.YAxis(), -self._rotTheta) *
            self._YZUpInvMatrix *
            Gf.Matrix4d().SetTranslate(self.center))
        self._camera.focusDistance = self.dist

        self._cameraTransformDirty = False

    def _pullFromCameraTransform(self):
        """
        Updates parameters (center, rotTheta, etc.) from the camera transform.
        """
        # reads the transform set on the camera and updates all the other
        # parameters.  This is the inverse of _pushToCameraTransform
        cam_transform = self._camera.transform
        dist = self._camera.focusDistance
        frustum = self._camera.frustum
        cam_pos = frustum.position
        cam_axis = frustum.ComputeViewDirection()

        # Compute translational parts
        self._dist = dist
        self._selSize = dist / 10.0
        self._center = cam_pos + dist * cam_axis

        # self._YZUpMatrix influences the behavior about how the
        # FreeCamera will tumble. It is the identity or a rotation about the
        # x-Axis.

        # Compute rotational part
        transform = cam_transform * self._YZUpMatrix
        transform.Orthonormalize()
        rotation = transform.ExtractRotation()

        # Decompose and set angles
        self._rotTheta, self._rotPhi, self._rotPsi = -rotation.Decompose(
            Gf.Vec3d.YAxis(), Gf.Vec3d.XAxis(), Gf.Vec3d.ZAxis())

        self._cameraTransformDirty = True


    def _rangeOfBoxAlongRay(self, camRay, bbox, debugClipping=False):
        maxDist = -float('inf')
        minDist = float('inf')
        boxRange = bbox.GetRange()
        boxXform = bbox.GetMatrix()
        for i in range (8):
            # for each corner of the bounding box, transform to world
            # space and project
            point = boxXform.Transform(boxRange.GetCorner(i))
            pointDist = camRay.FindClosestPoint(point)[1]

            # find the projection of that point of the camera ray
            # and find the farthest and closest point.
            if pointDist > maxDist:
                maxDist = pointDist
            if pointDist < minDist:
                minDist = pointDist

        if debugClipping:
            print("Projected bounds near/far: %f, %f" % (minDist, maxDist))

        # if part of the bbox is behind the ray origin (i.e. camera),
        # minDist will be zero, which is not usefull as a near-clip
        # distance.  In this case, give it a non-zero value that is
        # responsive to the actual scale of the scene.  If the scene
        # is fully in front of the camera, instead reduce minDist by a
        # bit so that geometry at exactly the edge of the bounds won't
        # be clipped - do the same for maxDist, also!
        if minDist == 0:
            minDist = min(FreeCamera.defaultNear,
                          (maxDist - minDist) *
                          FreeCamera.maxAllowableSceneClipFactor)
        else:
            minDist *= 0.99
        maxDist *= 1.01

        if debugClipping:
            print("Contracted bounds near/far: %f, %f" % (minDist, maxDist))

        return minDist, maxDist

    def setClippingPlanes(self, stageBBox):
        '''Computes and sets automatic clipping plane distances using the
           camera's position and orientation, the bouding box
           surrounding the stage, and the distance to the closest rendered
           object in the central view of the camera (closestVisibleDist).

           If either of the "override" clipping attributes are not None,
           we use those instead'''

        debugClipping = Tf.Debug.IsDebugSymbolNameEnabled(DEBUG_CLIPPING)

        # If the scene bounding box is empty, or we are fully on manual
        # override, then just initialize to defaults.
        if stageBBox.GetRange().IsEmpty() or \
               (self._overrideNear and self._overrideFar) :
            computedNear, computedFar = FreeCamera.defaultNear, FreeCamera.defaultFar
        else:
            # The problem: We want to include in the camera frustum all the
            # geometry the viewer should be able to see, i.e. everything within
            # the inifinite frustum starting at distance epsilon from the
            # camera itself.  However, the further the imageable geometry is
            # from the near-clipping plane, the less depth precision we will
            # have to resolve nearly colinear/incident polygons (which we get
            # especially with any doubleSided geometry).  We can run into such
            # situations astonishingly easily with large sets when we are
            # focussing in on just a part of a set that spans 10^5 units or
            # more.
            #
            # Our solution: Begin by projecting the endpoints of the imageable
            # world's bounds onto the ray piercing the center of the camera
            # frustum, and take the near/far clipping distances from its
            # extent, clamping at a positive value for near.  To address the
            # z-buffer precision issue, we rely on someone having told us how
            # close the closest imageable geometry actually is to the camera,
            # by having called setClosestVisibleDistFromPoint(). This gives us
            # the most liberal near distance we can use and not clip the
            # geometry we are looking at.  We actually choose some fraction of
            # that distance instead, because we do not expect the someone to
            # recompute the closest point with every camera manipulation, as
            # it can be expensive (we do emit signalFrustumChanged to notify
            # them, however).  We only use this if the current range of the
            # bbox-based frustum will have precision issues.
            frustum = self._camera.frustum
            camPos = frustum.position

            camRay = Gf.Ray(camPos, frustum.ComputeViewDirection())
            computedNear, computedFar = self._rangeOfBoxAlongRay(camRay,
                                                                 stageBBox,
                                                                 debugClipping)

            precisionNear = computedFar / FreeCamera.maxGoodZResolution

            if debugClipping:
                print("Proposed near for precision: {}, closestDist: {}"\
                    .format(precisionNear, self._closestVisibleDist))
            if self._closestVisibleDist:
                # Because of our concern about orbit/truck causing
                # clipping, make sure we don't go closer than half the
                # distance to the closest visible point
                halfClose = self._closestVisibleDist / 2.0

                if self._closestVisibleDist < self._lastFramedClosestDist:
                    # This can happen if we have zoomed in closer since
                    # the last time setClosestVisibleDistFromPoint() was called.
                    # Clamp to precisionNear, which gives a balance between
                    # clipping as we zoom in, vs bad z-fighting as we zoom in.
                    # See AdjustDistance() for comment about better solution.
                    halfClose = max(precisionNear, halfClose, computedNear)
                    if debugClipping:
                        print("ADJUSTING: Accounting for zoom-in")

                if halfClose < computedNear:
                    # If there's stuff very very close to the camera, it
                    # may have been clipped by computedNear.  Get it back!
                    computedNear = halfClose
                    if debugClipping:
                        print("ADJUSTING: closestDist was closer than bboxNear")
                elif precisionNear > computedNear:
                    computedNear = min((precisionNear + halfClose) / 2.0,
                                       halfClose)
                    if debugClipping:
                        print("ADJUSTING: gaining precision by pushing out")

        near = self._overrideNear or computedNear
        far  = self._overrideFar  or computedFar
        # Make sure far is greater than near
        far = max(near+1, far)

        if debugClipping:
            print("***Final Near/Far: {}, {}".format(near, far))

        self._camera.clippingRange = Gf.Range1f(near, far)

    def computeGfCamera(self, stageBBox, autoClip=False):
        """Makes sure the FreeCamera's computed parameters are up-to-date, and
        returns the GfCamera object.  If 'autoClip' is True, then compute
        "optimal" positions for the near/far clipping planes based on the
        current closestVisibleDist, in order to maximize Z-buffer resolution"""
        self._pushToCameraTransform()
        if autoClip:
            self.setClippingPlanes(stageBBox)
        else:
            self.resetClippingPlanes()
        return self._camera

    def resetClippingPlanes(self):
        """Set near and far back to their uncomputed defaults."""
        near = self._overrideNear or FreeCamera.defaultNear
        far  = self._overrideFar  or FreeCamera.defaultFar
        self._camera.clippingRange = Gf.Range1f(near, far)


    def frameSelection(self, selBBox, frameFit):
        # needs to be recomputed
        self._closestVisibleDist = None

        self.center = selBBox.ComputeCentroid()
        selRange = selBBox.ComputeAlignedRange()
        self._selSize = max(*selRange.GetSize())
        if self.orthographic:
            self.fov = self._selSize * frameFit
            self.dist = self._selSize + FreeCamera.defaultNear
        else:
            halfFov = self.fov*0.5 or 0.5 # don't divide by zero
            lengthToFit = self._selSize * frameFit * 0.5
            self.dist = lengthToFit / atan(rad(halfFov))
            # Very small objects that fill out their bounding boxes (like cubes)
            # may well pierce our 1 unit default near-clipping plane. Make sure
            # that doesn't happen.
            if self.dist < FreeCamera.defaultNear + self._selSize * 0.5:
                self.dist = FreeCamera.defaultNear + lengthToFit

    def setClosestVisibleDistFromPoint(self, point):
        frustum = self._camera.frustum
        camPos = frustum.position
        camRay = Gf.Ray(camPos, frustum.ComputeViewDirection())
        self._closestVisibleDist = camRay.FindClosestPoint(point)[1]
        self._lastFramedDist = self.dist
        self._lastFramedClosestDist = self._closestVisibleDist

        if Tf.Debug.IsDebugSymbolNameEnabled(DEBUG_CLIPPING):
            print("Resetting closest distance to {}; CameraPos: {}, closestPoint: {}".format(self._closestVisibleDist, camPos, point))

    def ComputePixelsToWorldFactor(self, viewportHeight):
        '''Computes the ratio that converts pixel distance into world units.

        It treats the pixel distances as if they were projected to a plane going
        through the camera center.'''
        self._pushToCameraTransform()
        if self.orthographic:
            return self.fov / viewportHeight
        else:
            frustumHeight = self._camera.frustum.window.GetSize()[1]
            return frustumHeight * self._dist / viewportHeight

    def Tumble(self, dTheta, dPhi):
        ''' Tumbles the camera around the center point by (dTheta, dPhi) degrees. '''
        self._rotTheta += dTheta
        self._rotPhi += dPhi
        self._cameraTransformDirty = True
        self.signalFrustumChanged.emit()

    def AdjustDistance(self, scaleFactor):
        '''Scales the distance of the freeCamera from it's center typically by
        scaleFactor unless it puts the camera into a "stuck" state.'''

        # When dist gets very small, you can get stuck and not be able to
        # zoom back out, if you just keep multiplying.  Switch to addition
        # in that case, choosing an incr that works for the scale of the
        # framed geometry.
        if scaleFactor > 1 and self.dist < 2:
            selBasedIncr = self._selSize / 25.0
            scaleFactor -= 1.0
            self.dist += min(selBasedIncr, scaleFactor)
        else:
            self.dist *= scaleFactor

        # Make use of our knowledge that we are changing distance to camera
        # to also adjust _closestVisibleDist to keep it useful.  Make sure
        # not to recede farther than the last *computed* closeDist, since that
        # will generally cause unwanted clipping of close objects.
        # XXX:  This heuristic does a good job of preventing undesirable
        # clipping as we zoom in and out, but sacrifices the z-buffer
        # precision we worked hard to get.  If Hd/UsdImaging could cheaply
        # provide us with the closest-point from the last-rendered image,
        # we could use it safely here to update _closestVisibleDist much
        # more accurately than this calculation.
        if self._closestVisibleDist:
            if self.dist > self._lastFramedDist:
                self._closestVisibleDist = self._lastFramedClosestDist
            else:
                self._closestVisibleDist = \
                    self._lastFramedClosestDist - \
                    self._lastFramedDist + \
                    self.dist

    def Truck(self, deltaRight, deltaUp):
        ''' Moves the camera by (deltaRight, deltaUp) in worldspace coordinates. 

        This is similar to a camera Truck/Pedestal.
        '''
        # need to update the camera transform before we access the frustum
        self._pushToCameraTransform()
        frustum = self._camera.frustum
        cam_up = frustum.ComputeUpVector()
        cam_right = Gf.Cross(frustum.ComputeViewDirection(), cam_up)
        self._center += (deltaRight * cam_right + deltaUp * cam_up)
        self._cameraTransformDirty = True
        self.signalFrustumChanged.emit()

    def PanTilt(self, dPan, dTilt):
        ''' Rotates the camera around the current camera base (approx. the film
        plane).  Both parameters are in degrees.

        This moves the center point that we normally tumble around.

        This is similar to a camera Pan/Tilt.
        '''
        self._camera.transform = (
                Gf.Matrix4d(1.0).SetRotate(Gf.Rotation(Gf.Vec3d.XAxis(), dTilt)) *
                Gf.Matrix4d(1.0).SetRotate(Gf.Rotation(Gf.Vec3d.YAxis(), dPan)) *
                self._camera.transform)
        self._pullFromCameraTransform()

        # When we Pan/Tilt, we don't want to roll the camera so we just zero it
        # out here.
        self._rotPsi = 0.0

        self._cameraTransformDirty = True
        self.signalFrustumChanged.emit()

    def Walk(self, dForward, dRight):
        ''' Specialized camera movement that moves it on the "horizontal" plane
        '''
        # need to update the camera transform before we access the frustum
        self._pushToCameraTransform()
        frustum = self._camera.frustum
        cam_up = frustum.ComputeUpVector().GetNormalized()
        cam_forward = frustum.ComputeViewDirection().GetNormalized()
        cam_right = Gf.Cross(cam_forward, cam_up)
        delta = dForward * cam_forward + dRight * cam_right
        self._center += delta
        self._cameraTransformDirty = True
        self.signalFrustumChanged.emit()

    @staticmethod
    def FromGfCamera(gfCamera, isZUp):
        self = FreeCamera(isZUp)
        self._camera = gfCamera
        self._pullFromCameraTransform()
        return self

    @property
    def rotTheta(self):
        return self._rotTheta

    @rotTheta.setter
    def rotTheta(self, value):
        self._rotTheta = value
        self._cameraTransformDirty = True
        self.signalFrustumChanged.emit()

    @property
    def rotPhi(self):
        return self._rotPhi

    @rotPhi.setter
    def rotPhi(self, value):
        self._rotPhi = value
        self._cameraTransformDirty = True
        self.signalFrustumChanged.emit()

    @property
    def center(self):
        return self._center

    @center.setter
    def center(self, value):
        self._center = value
        self._cameraTransformDirty = True
        self.signalFrustumChanged.emit()

    @property
    def dist(self):
        return self._dist

    @dist.setter
    def dist(self, value):
        self._dist = value
        self._cameraTransformDirty = True
        self.signalFrustumChanged.emit()

    @property
    def orthographic(self):
        return self._camera.projection == Gf.Camera.Orthographic

    @orthographic.setter
    def orthographic(self, orthographic):
        if orthographic:
            self._camera.projection = Gf.Camera.Orthographic
        else:
            self._camera.projection = Gf.Camera.Perspective
        self.signalFrustumChanged.emit()
        self.signalFrustumSettingsChanged.emit()

    @property
    def fov(self):
        """The vertical field of view, in degrees, for perspective cameras. 
        For orthographic cameras fov is the height of the view frustum, in 
        world units.
        """
        if self._camera.projection == Gf.Camera.Perspective:
            return self._camera.GetFieldOfView(Gf.Camera.FOVVertical)
        else:
            return (self._camera.verticalAperture * Gf.Camera.APERTURE_UNIT)

    @fov.setter
    def fov(self, value):
        if self._camera.projection == Gf.Camera.Perspective:
            self._camera.SetPerspectiveFromAspectRatioAndFieldOfView(
                self._camera.aspectRatio, value, Gf.Camera.FOVVertical)
        else:
            self._camera.SetOrthographicFromAspectRatioAndSize(
                self._camera.aspectRatio, value, Gf.Camera.FOVVertical)
        self.signalFrustumChanged.emit()
        self.signalFrustumSettingsChanged.emit()

    @property
    def aspectRatio(self):
        return self._camera.aspectRatio

    @aspectRatio.setter
    def aspectRatio(self, value):
        """Sets the aspect ratio by adjusting the horizontal aperture."""
        self.horizontalAperture = value * self.verticalAperture

    @property
    def horizontalAperture(self):
        return self._camera.horizontalAperture

    @horizontalAperture.setter
    def horizontalAperture(self, value):
        self._camera.horizontalAperture = value
        self.signalFrustumChanged.emit()
        self.signalFrustumSettingsChanged.emit()

    @property
    def verticalAperture(self):
        return self._camera.verticalAperture

    @verticalAperture.setter
    def verticalAperture(self, value):
        self._camera.verticalAperture = value
        self.signalFrustumChanged.emit()
        self.signalFrustumSettingsChanged.emit()
    
    @property
    def focalLength(self):
        return self._camera.focalLength

    @focalLength.setter
    def focalLength(self, value):
        self._camera.focalLength = value
        self.signalFrustumChanged.emit()
        self.signalFrustumSettingsChanged.emit()

    @property
    def near(self):
        return self._camera.clippingRange.min


    @property
    def far(self):
        return self._camera.clippingRange.max

    # no setters for near and far - one must set overrideNear/Far instead
    @property
    def overrideNear(self):
        return self._overrideNear

    @overrideNear.setter
    def overrideNear(self, value):
        """To remove the override, set to None"""
        self._overrideNear = value

    @property
    def overrideFar(self):
        return self._overrideFar

    @overrideFar.setter
    def overrideFar(self, value):
        """To remove the override, set to None"""
        self._overrideFar = value



#
# Copyright 2016 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#

from pxr import UsdGeom, UsdShade, UsdSemantics
from pxr.UsdUtils.constantsGroup import ConstantsGroup


class ComputedPropertyNames(ConstantsGroup):
    """Names of all available computed properties."""
    WORLD_BBOX              = "World Bounding Box"
    LOCAL_WORLD_XFORM       = "Local to World Xform"
    RESOLVED_PREVIEW_MATERIAL = "Resolved Preview Material"
    RESOLVED_FULL_MATERIAL = "Resolved Full Material"
    RESOLVED_LABELS        = "Resolved Labels"

#
# Edit the following to alter the set of custom attributes.
#
# Every entry should be an object derived from CustomAttribute,
# defined below.
#
def _GetCustomAttributes(currentPrim, rootDataModel):
    currentPrimIsImageable = currentPrim.IsA(UsdGeom.Imageable)
    
    # If the currentPrim is imageable or if it is a typeless def, it 
    # participates in imageable computations.
    currentPrimGetsImageableComputations = currentPrim.IsA(UsdGeom.Imageable) \
            or not currentPrim.GetTypeName()
        
    if currentPrimGetsImageableComputations:
        return [BoundingBoxAttribute(currentPrim, rootDataModel),
                LocalToWorldXformAttribute(currentPrim, 
                                           rootDataModel),
                ResolvedPreviewMaterial(currentPrim, rootDataModel),
                ResolvedFullMaterial(currentPrim, rootDataModel),
                ResolvedLabelsAttribute(currentPrim, rootDataModel),
        ]
    return [ResolvedLabelsAttribute(currentPrim, rootDataModel)]

#
# The base class for per-prim custom attributes.
#
class CustomAttribute:
    def __init__(self, currentPrim, rootDataModel):
        self._currentPrim = currentPrim
        self._rootDataModel = rootDataModel

    def IsVisible(self):
        return True

    # GetName function to match UsdAttribute API
    def GetName(self):
        return ""

    # Get function to match UsdAttribute API
    def Get(self, frame):
        return ""

    # convenience function to make this look more like a UsdAttribute
    def GetTypeName(self):
        return ""

    # GetPrimPath function to match UsdAttribute API
    def GetPrimPath(self):
        return self._currentPrim.GetPath()

#
# Displays the bounding box of a prim
#
class BoundingBoxAttribute(CustomAttribute):
    def __init__(self, currentPrim, rootDataModel):
        CustomAttribute.__init__(self, currentPrim, rootDataModel)

    def GetName(self):
        return ComputedPropertyNames.WORLD_BBOX

    def Get(self, frame):
        try:
            bbox = self._rootDataModel.computeWorldBound(self._currentPrim)
            bbox = bbox.ComputeAlignedRange()
        except RuntimeError as err:
            bbox = "Invalid: " + str(err)

        return bbox

#
# Displays the Local to world xform of a prim
#
class LocalToWorldXformAttribute(CustomAttribute):
    def __init__(self, currentPrim, rootDataModel):
        CustomAttribute.__init__(self, currentPrim, rootDataModel)

    def GetName(self):
        return ComputedPropertyNames.LOCAL_WORLD_XFORM

    def Get(self, frame):
        try:
            pwt = self._rootDataModel.getLocalToWorldTransform(self._currentPrim)
        except RuntimeError as err:
            pwt = "Invalid: " + str(err)

        return pwt

class ResolvedBoundMaterial(CustomAttribute):
    def __init__(self, currentPrim, rootDataModel, purpose):
        CustomAttribute.__init__(self, currentPrim, rootDataModel)
        self._purpose = purpose

    def GetName(self):
        if self._purpose == UsdShade.Tokens.full:
            return ComputedPropertyNames.RESOLVED_FULL_MATERIAL
        elif self._purpose == UsdShade.Tokens.preview:
            return ComputedPropertyNames.RESOLVED_PREVIEW_MATERIAL
        else:
            raise ValueError("Invalid purpose '{}'.".format(self._purpose))

    def Get(self, frame):
        try:
            (boundMaterial, bindingRel) = \
                self._rootDataModel.computeBoundMaterial(self._currentPrim, 
                        self._purpose)
            boundMatPath = boundMaterial.GetPrim().GetPath() if boundMaterial \
                else "<unbound>"
        except RuntimeError as err:
            boundMatPath = "Invalid: " + str(err)
        return boundMatPath

class ResolvedFullMaterial(ResolvedBoundMaterial):
    def __init__(self, currentPrim, rootDataModel):
        ResolvedBoundMaterial.__init__(self, currentPrim, rootDataModel, 
                UsdShade.Tokens.full)

class ResolvedPreviewMaterial(ResolvedBoundMaterial):
    def __init__(self, currentPrim, rootDataModel):
        ResolvedBoundMaterial.__init__(self, currentPrim, rootDataModel, 
                UsdShade.Tokens.preview)

#
# Displays a prim's inherited labels
#
class ResolvedLabelsAttribute(CustomAttribute):
    def GetName(self):
        return ComputedPropertyNames.RESOLVED_LABELS

    def Get(self, frame):
        try:
            labels = self._rootDataModel.getResolvedLabels(
                    self._currentPrim, frame)
        except RuntimeError as err:
            labels = "Invalid: " + str(err)

        return labels


class ComputedPropertyFactory:
    """Creates computed properties."""

    def __init__(self, rootDataModel):

        self._rootDataModel = rootDataModel

    def getComputedProperty(self, prim, propName):
        """Create a new computed property from a prim and property name."""

        if propName == ComputedPropertyNames.WORLD_BBOX:
            return BoundingBoxAttribute(prim, self._rootDataModel)
        elif propName == ComputedPropertyNames.LOCAL_WORLD_XFORM:
            return LocalToWorldXformAttribute(prim, self._rootDataModel)
        elif propName == ComputedPropertyNames.RESOLVED_FULL_MATERIAL:
            return ResolvedFullMaterial(prim, self._rootDataModel)
        elif propName == ComputedPropertyNames.RESOLVED_PREVIEW_MATERIAL:
            return ResolvedPreviewMaterial(prim, self._rootDataModel)
        elif propName == ComputedPropertyNames.RESOLVED_LABELS:
            return ResolvedLabelsAttribute(prim, self._rootDataModel)
        else:
            raise ValueError("Cannot create computed property '{}'.".format(
                propName))



#
# Copyright 2024 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#
"""These utilities parse usdview timing output into tuples of
(success, identifier, timing).
"""
import re

make_re = lambda s: re.compile(r'(Time\ to\ )' + s + '(.*s)')

# `usdview --timing` metrics
METRICS = [
    ("configure_and_load_plugins",  make_re(r'(configure\ and\ load\ plugins:)')),
    ("open_stage", make_re(r'(open\ stage.*:)')),
    ("reset_prim_browser", make_re(r'(reset\ Prim\ Browser\ to.*:)')),
    ("bring_up_the_ui", make_re(r'(bring\ up\ the\ UI:)')),
    ("create_first_image", make_re(r'(create\ first\ image:)')),
    ("shut_down_hydra", make_re(r'(shut\ down\ Hydra:)')),
    ("close_stage", make_re(r'(close\ stage.*:)')),
    ("tear_down_the_ui", make_re(r'(tear\ down\ the\ UI:)')),
    ("open_and_close_usdview", make_re(r'(open\ and\ close\ usdview:)')),
]


def nameToMetricIdentifier(desc):
    """
    From a metric name, return a standardized identifier
    following an "underscore instead of spaces" convention.
    """
    return desc.lower().replace(" ", "_")


def parseTiming(line):
    """
    Parses tuple of (success, metric_identifier, time) from the given
    input line based on the identifier, regex combinations in METRICS.
    """
    for ident, re_expr in METRICS:
        match = re_expr.match(line)
        if match:
            groups = match.groups()
            time = float(groups[2].strip()[:-1])
            return (True, ident, time)
    return (False, None, None)


def parseTimingGeneric(metricExpressions, line):
    """
    Produces tuple of (success, metric_identifier, time) from the given
    input line based on regexes provided by metricExpressions.
    """
    metrics = [
        make_re('{profile}:'.format(profile=profile))
        for profile in metricExpressions
    ]

    for e in metrics:
        match = e.match(line)
        if match:
            groups = match.groups()
            metricName = groups[1].strip()
            ident = nameToMetricIdentifier(metricName)
            time = float(groups[2].strip()[:-1])
            return (True, ident, time)

    return (False, None, None)



#
# Copyright 2017 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#
from .qt import QtWidgets

from .primLegendUI import Ui_PrimLegend

from .common import (UIPrimTypeColors,
                     ColorizeLabelText, BoldenLabelText, ItalicizeLabelText)

class PrimLegend(QtWidgets.QWidget):
    def __init__(self, parent):
        QtWidgets.QWidget.__init__(self, parent)
        self._ui = Ui_PrimLegend()
        self._ui.setupUi(self)

        # start out in a minimized/hidden state
        self.setMaximumHeight(0)
        self._isMinimized = True

        graphicsScene = QtWidgets.QGraphicsScene()

        # Set colors
        self._ui.primLegendColorHasArcs.setScene(graphicsScene)
        self._ui.primLegendColorNormal.setScene(graphicsScene)
        self._ui.primLegendColorInstance.setScene(graphicsScene)
        self._ui.primLegendColorPrototype.setScene(graphicsScene)

        self._ui.primLegendColorHasArcs.setForegroundBrush(UIPrimTypeColors.HAS_ARCS)
        self._ui.primLegendColorNormal.setForegroundBrush(UIPrimTypeColors.NORMAL)
        self._ui.primLegendColorInstance.setForegroundBrush(UIPrimTypeColors.INSTANCE)
        self._ui.primLegendColorPrototype.setForegroundBrush(UIPrimTypeColors.PROTOTYPE)

        legendTextUpdate = lambda t, c: (('<font color=\"%s\">' % c.color().name())
                                             + t.text() + '</font>')

        normalLegend = self._ui.primLegendLabelNormal
        normalLegend.setText(legendTextUpdate(normalLegend, UIPrimTypeColors.NORMAL))

        prototypeLegend = self._ui.primLegendLabelPrototype
        prototypeLegend.setText(legendTextUpdate(prototypeLegend, UIPrimTypeColors.PROTOTYPE))

        instanceLegend = self._ui.primLegendLabelInstance
        instanceLegend.setText(legendTextUpdate(instanceLegend, UIPrimTypeColors.INSTANCE))

        hasArcsLegend = self._ui.primLegendLabelHasArcs
        hasArcsLegend.setText(legendTextUpdate(hasArcsLegend, UIPrimTypeColors.HAS_ARCS))

        undefinedFontLegend = self._ui.primLegendLabelFontsUndefined
        undefinedFontLegend.setText(ItalicizeLabelText(undefinedFontLegend.text(),
                                                       undefinedFontLegend.text()))

        definedFontLegend = self._ui.primLegendLabelFontsDefined
        definedFontLegend.setText(BoldenLabelText(definedFontLegend.text(),
                                                  definedFontLegend.text()))

        # Set three individual colors in the text line to indicate
        # the dimmed version of each primary prim color
        dimmedLegend = self._ui.primLegendLabelDimmed
        dimmedLegendText = dimmedLegend.text()
        dimmedLegendText = ColorizeLabelText(
            dimmedLegendText, "Dimmed colors", 148, 105, 30)
        dimmedLegendText = ColorizeLabelText(
            dimmedLegendText, "denote", 78, 91, 145)
        dimmedLegendText = ColorizeLabelText(
            dimmedLegendText, "inactive prims", 151, 151, 151)
        dimmedLegend.setText(dimmedLegendText)

    def IsMinimized(self):
        return self._isMinimized

    def ToggleMinimized(self):
        self._isMinimized = not self._isMinimized

    def GetHeight(self):
        return self.height()

    def GetResetHeight(self):
        return self.sizeHint().height()



__all__ = ['Tf', 'Gf', 'Trace', 'Work', 'Plug', 'Vt', 'Ts', 'Ar', 'Kind', 'Sdf', 'Sdr', 'Pcp', 'Usd', 'UsdGeom', 'UsdVol', 'UsdMedia', 'UsdShade', 'UsdLod', 'UsdLux', 'UsdProc', 'UsdProfiles', 'UsdRender', 'UsdHydra', 'UsdRi', 'UsdSemantics', 'UsdSkel', 'UsdUI', 'UsdUtils', 'UsdPhysics', 'UsdAbc', 'UsdDraco', 'UsdMtlx', 'ExecIr', 'UsdValidation', 'CameraUtil', 'PxOsd', 'GeomUtil', 'Glf', 'UsdImagingGL', 'UsdAppUtils', 'Usdviewq', 'UsdBakeMtlx']



#
# Copyright 2016 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#
"""
Tf -- Tools Foundation
"""

# Type to help handle DLL import paths on Windows with python interpreters v3.8
# and newer. These interpreters don't search for DLLs in the path anymore, you
# have to provide a path explicitly. This re-enables path searching for USD 
# dependency libraries
import platform, sys
if sys.version_info >= (3, 8) and platform.system() == "Windows":
    import contextlib

    @contextlib.contextmanager
    def WindowsImportWrapper():
        import os
        dirs = []
        import_paths = os.getenv('PXR_USD_WINDOWS_DLL_PATH')
        if import_paths is None:
            import_paths = os.getenv('PATH', '')
        # the underlying windows API call, AddDllDirectory, states that:
        #
        # > If AddDllDirectory is used to add more than one directory to the
        # > process DLL search path, the order in which those directories are
        # > searched is unspecified.
        #
        # https://learn.microsoft.com/en-us/windows/win32/api/libloaderapi/nf-libloaderapi-adddlldirectory
        #
        # However, in practice, it seems that the most-recently-added ones
        # take precedence - so, reverse the order of entries in PATH to give
        # it the same precedence
        #
        # Note that we have a test (testTfPyDllLink) to alert us if this
        # undefined behavior changes.
        for path in reversed(import_paths.split(os.pathsep)):
            # Calling add_dll_directory raises an exception if paths don't
            # exist, or if you pass in dot
            if os.path.exists(path) and path != '.':
                abs_path = os.path.abspath(path)
                dirs.append(os.add_dll_directory(abs_path))
        # This block guarantees we clear the dll directories if an exception
        # is raised in the with block.
        try:
            yield
        finally:
            for dll_dir in dirs:
                dll_dir.close()
        del os
    del contextlib
else:
    class WindowsImportWrapper(object):
        def __enter__(self):
            pass
        def __exit__(self, exc_type, ex_val, exc_tb):
            pass
del platform, sys


def PreparePythonModule(moduleName=None):
    """Prepare an extension module at import time.  This will import the
    Python module associated with the caller's module (e.g. '_tf' for 'pxr.Tf')
    or the module with the specified moduleName and copy its contents into
    the caller's local namespace.

    Generally, this should only be called by the __init__.py script for a module
    upon loading a boost python module (generally '_libName.so')."""
    import importlib
    import inspect
    frame = inspect.currentframe().f_back
    try:
        f_locals = frame.f_locals

        # If an explicit moduleName is not supplied, construct it from the
        # caller's module name, like "pxr.Tf", and our naming conventions,
        # which results in "_tf".
        if moduleName is None:
            moduleName = f_locals["__name__"].split(".")[-1]
            moduleName = "_" + moduleName[0].lower() + moduleName[1:]

        with WindowsImportWrapper():
            module = importlib.import_module(
                    "." + moduleName, f_locals["__name__"])

        PrepareModule(module, f_locals)
        try:
            del f_locals[moduleName]
        except KeyError:
            pass

        try:
            module = importlib.import_module(".__DOC", f_locals["__name__"])
            module.Execute(f_locals)
            try:
                del f_locals["__DOC"]
            except KeyError:
                pass
        except Exception:
            pass

    finally:
        del frame

def PrepareModule(module, result):
    """PrepareModule(module, result) -- Prepare an extension module at import
    time.  Generally, this should only be called by the __init__.py script for a
    module upon loading a boost python module (generally '_libName.so')."""
    # inject into result.
    ignore = frozenset(['__name__', '__package__', '__builtins__',
                        '__doc__', '__file__', '__path__'])
    newModuleName = result.get('__name__')

    for key, value in module.__dict__.items():
        if not key in ignore:
            result[key] = value

            # Lie about the module from which value came.
            if newModuleName and hasattr(value, '__module__'):
                try:
                    setattr(value, '__module__', newModuleName)
                except AttributeError as e:
                    # The __module__ attribute of Boost.Python.function
                    # objects is not writable, so we get this exception
                    # a lot.  Just ignore it.  We're really only concerned
                    # about the data objects like enum values and such.
                    #
                    pass

def GetCodeLocation(framesUp):
    """Returns a tuple (moduleName, functionName, fileName, lineNo).

    To trace the current location of python execution, use GetCodeLocation().
    By default, the information is returned at the current stack-frame; thus

        info = GetCodeLocation()

    will return information about the line that GetCodeLocation() was called 
    from. One can write:

        def genericDebugFacility():
            info = GetCodeLocation(1)
            # print out data


        def someCode():
            ...
            if bad:
                genericDebugFacility()

    and genericDebugFacility() will get information associated with its caller, 
    i.e. the function someCode()."""
    import sys
    f_back = sys._getframe(framesUp).f_back
    return (f_back.f_globals.get('__name__', '<unknown>'),
            f_back.f_code.co_name,
            f_back.f_code.co_filename,
            f_back.f_lineno)

PreparePythonModule()

# Need to provide an exception type that tf errors will show up as.
class ErrorException(RuntimeError):
    def __init__(self, *args):
        RuntimeError.__init__(self, *args)
        self.__TfException = True

    def __str__(self):
        return '\n\t' + '\n\t'.join([str(e) for e in self.args])
__SetErrorExceptionClass(ErrorException)

def Warn(msg, template=""):
    """Issue a warning via the TfDiagnostic system.

    At this time, template is ignored.
    """
    codeInfo = GetCodeLocation(framesUp=1)
    _Warn(msg, codeInfo[0], codeInfo[1], codeInfo[2], codeInfo[3])
    
def Status(msg, verbose=True):
    """Issues a status update to the Tf diagnostic system.

    If verbose is True (the default) then information about where in the code
    the status update was issued from is included.
    """
    if verbose:
        codeInfo = GetCodeLocation(framesUp=1)
        _Status(msg, codeInfo[0], codeInfo[1], codeInfo[2], codeInfo[3])
    else:
        _Status(msg)

def RaiseCodingError(msg):
    """Raise a coding error to the Tf Diagnostic system."""
    codeInfo = GetCodeLocation(framesUp=1)
    _RaiseCodingError(msg, codeInfo[0], codeInfo[1], codeInfo[2], codeInfo[3])

def RaiseRuntimeError(msg):
    """Raise a runtime error to the Tf Diagnostic system."""
    codeInfo = GetCodeLocation(framesUp=1)
    _RaiseRuntimeError(msg, codeInfo[0], codeInfo[1], codeInfo[2], codeInfo[3])

def Fatal(msg):
    """Raise a fatal error to the Tf Diagnostic system."""
    codeInfo = GetCodeLocation(framesUp=1)
    _Fatal(msg, codeInfo[0], codeInfo[1], codeInfo[2], codeInfo[3])


class NamedTemporaryFile(object):
    """A named temporary file which keeps the internal file handle closed. 
       A class which constructs a temporary file(that isn't open) on __enter__,
       provides its name as an attribute, and deletes it on __exit__. 
       
       Note: The constructor args for this object match those of 
       python's tempfile.mkstemp() function, and will have the same effect on
       the underlying file created."""

    def __init__(self, suffix='', prefix='', dir=None, text=False):
        # Note that we defer creation until the enter block to 
        # prevent users from unintentionally creating a bunch of
        # temp files that don't get cleaned up.
        self._args = (suffix,  prefix, dir, text)

    def __enter__(self):
        from tempfile import mkstemp
        from os import close
        
        fd, path = mkstemp(*self._args)
        close(fd) 

        # XXX: We currently only expose the name attribute
        # more can be added based on client needs in the future.
        self._name = path

        return self

    def __exit__(self, *args):
        import os
        os.remove(self.name)

    @property
    def name(self):
        """The path for the temporary file created.""" 
        return self._name

# Extend Tf.ScopeDescription using ContextDecorator so that it can also be
# used to decorate functions/methods.
import contextlib

class ScopeDescription(_ScopeDescription, contextlib.ContextDecorator):
    """This class lets us expose TfScopeDescription to python for use as a
    "context manager" object, for use with the 'with'-statement. For example:

        with Tf.ScopeDescription("Solving the halting problem"):
            # Code that solves the halting problem.

    It can also be used as a decorator for an entire function or method, for
    convenience:

        @Tf.ScopeDescription("Solving the halting problem");
        def SolveHaltingProblem():
            # Code that solves the halting problem.

    Note that the description passed to the decorator is an expression that is
    evaluated only once at declaration time. If a dynamic description is
    needed, use the with-statement pattern instead.
    """
    pass

class CatchAndRepostErrors(contextlib.ContextDecorator):
    """This context manager & decorator can be used to catch Tf.ErrorException
    and repost its held errors to the thread's Tf.Error list without raising a
    Python exception, halting unwinding.  For example, the following will raise
    a Tf.ErrorException containing the "test" Coding Error.  This will be
    caught by the CatchAndRepostErrors object, and the "test" Error will be
    re-posted to the thread's error list.  Then exception propagation will
    cease and ordinary control flow will continue.

        with Tf.CatchAndRepostErrors():
            Tf.RaiseCodingError('test')

    This can be useful to sidestep SIP's virtual function bindings from
    blocking exceptions, retaining Tf.Errors so they can propagate to callers.
    Here, the `paintGL()` virtual override can capture any Tf.ErrorExceptions
    raised during its execution, and by reposting the errors to the Tf.Error
    list, callers can observe them later.

        @Tf.CatchAndRepostErrors():
        def paintGL(self):
            self._renderer.Render()
    """

    def __enter__(self):
        return self

    def __exit__(self, excType, excValue, tb):
        return (RepostErrors(excValue)
                if isinstance(excValue, ErrorException)
                else False)

del contextlib



#
# Copyright 2018 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#

"""Python bindings for libSdr"""

from pxr import Tf
Tf.PreparePythonModule()
del Tf



#
# Copyright 2018 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#
from .qt import QtCore, QtWidgets, QtGui


class FrameSlider(QtWidgets.QSlider):
    """Custom QSlider class to allow scrubbing on left-click."""

    PAUSE_TIMER_INTERVAL = 500

    def __init__(self, parent=None):
        super(FrameSlider, self).__init__(parent=parent)
        # Create a mouse pause timer to trigger an update if the slider
        # scrubbing pauses.
        self._mousePauseTimer = QtCore.QTimer(self)
        self._mousePauseTimer.setInterval(self.PAUSE_TIMER_INTERVAL)
        self._mousePauseTimer.timeout.connect(self.mousePaused)

    def mousePaused(self):
        """Slot called when the slider scrubbing is paused."""
        self._mousePauseTimer.stop()
        self.valueChanged.emit(self.sliderPosition())

    def mousePressEvent(self, event):
        # This is a temporary solution that should revisited in future.
        #
        # The issue is that the current QStyle on the application
        # only allows the slider position to be set on a MiddleButton
        # MouseEvent.
        #
        # The correct solution is for us to create a QProxyStyle for the
        # application so we can edit the value of the
        # QStyle.StyleHint.SH_Slider_AbsoluteSetButtons property (to include
        # LeftButton). Unfortunately QProxyStyle is not yet available
        # in this version of PySide.
        #
        # Instead, we are forced to duplicate the MouseEvent as a
        # MiddleButton event. This creates the exact behavior we
        # want to see from the QSlider.
        styleHint = QtWidgets.QStyle.StyleHint.SH_Slider_AbsoluteSetButtons
        if self.style().styleHint(styleHint) == QtCore.Qt.MouseButton.MiddleButton:
            if event.button() == QtCore.Qt.MouseButton.LeftButton:
                event = QtGui.QMouseEvent(
                    event.type(),
                    event.pos(),
                    QtCore.Qt.MouseButton.MiddleButton,
                    QtCore.Qt.MouseButton.MiddleButton,
                    event.modifiers()
                )

        super(FrameSlider, self).mousePressEvent(event)

    def mouseMoveEvent(self, event):
        super(FrameSlider, self).mouseMoveEvent(event)

        # Start the pause timer if tracking is disabled.
        if not self.hasTracking():
            self._mousePauseTimer.start()

    def mouseReleaseEvent(self, event):
        # Stop the pause timer.
        self._mousePauseTimer.stop()

        super(FrameSlider, self).mouseReleaseEvent(event)



#
# Copyright 2018 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#
from .qt import QtCore, QtGui, QtWidgets
from pxr.UsdUtils.constantsGroup import ConstantsGroup
from pxr import Sdf, Usd, UsdGeom
from .primViewItem import PrimViewColumnIndex, PrimViewItem
from .common import PrintWarning, Timer, UIPrimTreeColors, KeyboardShortcuts

def _GetPropertySpecInSessionLayer(usdAttribute):
    propertyStack = usdAttribute.GetPropertyStack(Usd.TimeCode.Default())
    if len(propertyStack) > 0:
        stageSessionLayer = usdAttribute.GetStage().GetSessionLayer()
        return stageSessionLayer.GetPropertyAtPath(usdAttribute.GetPath())
    return None

# Function for getting the background color of the item for the delegates.
# Returns none if we only want the default paint method. 
def _GetBackgroundColor(item, option):
    mouseOver = option.state & QtWidgets.QStyle.StateFlag.State_MouseOver
    selected = option.state & QtWidgets.QStyle.StateFlag.State_Selected
    pressed = option.state & QtWidgets.QStyle.StateFlag.State_Sunken

    background = None

    if item.ancestorOfSelected:
        background = UIPrimTreeColors.ANCESTOR_OF_SELECTED
        if mouseOver:
            background = UIPrimTreeColors.ANCESTOR_OF_SELECTED_HOVER
        if selected:
            background = UIPrimTreeColors.SELECTED
            if mouseOver:
                background = UIPrimTreeColors.SELECTED_HOVER

    else:
        if not selected and not pressed and mouseOver:
            background = UIPrimTreeColors.UNSELECTED_HOVER

        if selected:
            background = UIPrimTreeColors.SELECTED
            if mouseOver:
                background = UIPrimTreeColors.SELECTED_HOVER

    return background

class DrawModes(ConstantsGroup):
    DEFAULT = "default"
    CARDS = "cards"
    BOUNDS = "bounds"
    ORIGIN = "origin"

class DrawModeComboBox(QtWidgets.QComboBox):
    """ Specialize from QComboBox, so that we can send a signal when the pop-up
        is hidden.
    """
    signalPopupHidden = QtCore.Signal()

    def __init__(self, parent=None):
        QtWidgets.QComboBox.__init__(self, parent)

    def hidePopup(self):
        QtWidgets.QComboBox.hidePopup(self)
        self.signalPopupHidden.emit()

class DrawModeWidget(QtWidgets.QWidget):
    """ This widget contains a combobox for selecting the draw mode and a 
        clear ('x') button for clearing an authored drawMode override in the 
        session layer. 
    """
    def __init__(self, primViewItem, refreshFunc, makeTimer,
                 parent=None):
        QtWidgets.QWidget.__init__(self, parent)

        self._primViewItem = primViewItem
        self._layout = QtWidgets.QHBoxLayout()
        self._layout.setSpacing(0)
        self._layout.setContentsMargins(0,0,0,0)
        self.setLayout(self._layout)
        
        self._comboBox = DrawModeComboBox(self)
        self._modelAPI = UsdGeom.ModelAPI(self._primViewItem.prim)
        # Reducing the width further causes the pop-up dialog to be trimmed 
        # and option text to be pruned.
        self._comboBox.setFixedWidth(100)
        self._comboBox.addItem(DrawModes.DEFAULT)
        self._comboBox.addItem(DrawModes.CARDS)
        self._comboBox.addItem(DrawModes.BOUNDS)
        self._comboBox.addItem(DrawModes.ORIGIN)
        self._layout.addWidget(self._comboBox)

        self._clearButton = QtWidgets.QToolButton(self)
        self._clearButton.setText('X')
        self._clearButton.setFixedSize(16, 16)
        retainSizePolicy = self._clearButton.sizePolicy()
        retainSizePolicy.setRetainSizeWhenHidden(True)
        self._clearButton.setSizePolicy(retainSizePolicy)
        self._layout.addWidget(self._clearButton)            
        
        self._currentDrawMode = None
        self.RefreshDrawMode()
        self._firstPaint = True

        self._refreshFunc = refreshFunc
        self._makeTimer = makeTimer

        self._comboBox.signalPopupHidden.connect(self._PopupHidden)
        self._comboBox.activated.connect(self._UpdateDrawMode)
        self._clearButton.clicked.connect(self._ClearDrawMode)

    def paintEvent(self, event):
        # Virtual override of the paintEvent method on QWidget.
        # Popup the combo box the first time the widget is drawn, since 
        # mouse-down in the column makes the widget appear.
        # 
        # An alternative approach would be to set a timer in a constructor to 
        # pop open the combo box at the end of the event loop, but it causes a 
        # noticeable flicker in the UI.
        if self._firstPaint:
            self._comboBox.showPopup()
            self._firstPaint = False

        # Invoke paintEvent on super class to do the actual painting of the 
        # widget.
        super(DrawModeWidget, self).paintEvent(event)

    def _ShouldHideClearButton(self):
        # Check if there's an authored value in the session layer.
        drawModeAttr = self._modelAPI.GetModelDrawModeAttr()
        
        if drawModeAttr:
            sessionSpec = _GetPropertySpecInSessionLayer(drawModeAttr)
            if sessionSpec and sessionSpec.HasDefaultValue():
                return False
        return True

    def RefreshDrawMode(self, currentDrawMode=None):
        self._currentDrawMode = currentDrawMode if currentDrawMode else \
                                self._modelAPI.ComputeModelDrawMode()
        self._comboBox.setCurrentText(self._currentDrawMode)

        clearButtonIsHidden = self._clearButton.isHidden()
        if self._ShouldHideClearButton():
            if not clearButtonIsHidden:
                self._clearButton.hide()
                self.update()
        else:
            if clearButtonIsHidden:
                self._clearButton.show()
                self.update()

    def _UpdateDrawMode(self):
        newDrawModeSelection = str(self._comboBox.currentText())
        currentDrawMode = self._modelAPI.ComputeModelDrawMode()
        if currentDrawMode != newDrawModeSelection:
            with self._makeTimer("change model:drawMode on <%s> to %s" % 
                             (self._modelAPI.GetPath(), newDrawModeSelection)):
                self._modelAPI = UsdGeom.ModelAPI.Apply(self._primViewItem.prim)
                self._modelAPI.CreateModelDrawModeAttr().Set(
                        newDrawModeSelection)

                self.RefreshDrawMode(currentDrawMode=newDrawModeSelection)

                # We need to redraw the scene to pick up updates to draw mode.
                self._refreshFunc(self._primViewItem)
        self._CloseEditorIfNoEdit()

    def _ClearDrawMode(self):
        with self._makeTimer("clear model:drawMode on <%s>" % 
                         self._modelAPI.GetPath()):
            drawModeAttr = self._modelAPI.GetModelDrawModeAttr()
            if drawModeAttr:
                sessionSpec = _GetPropertySpecInSessionLayer(drawModeAttr)
                if sessionSpec:
                    self._primViewItem.drawModeWidget = None
                    self._primViewItem.treeWidget().closePersistentEditor(
                            self._primViewItem, PrimViewColumnIndex.DRAWMODE)

                    sessionSpec.ClearDefaultValue()
                    sessionSpec.layer.ScheduleRemoveIfInert(sessionSpec)
                    self._refreshFunc(self._primViewItem)
                else:
                    PrintWarning(self._modelAPI.GetPath(), "Failed to get "
                        "session layer spec for the model:drawMode attribute")
                    return
            else:
                PrintWarning(self._modelAPI.GetPath(), "Failed to get "
                    "model:drawMode attribute")
                return

    def _CloseEditorIfNoEdit(self):
        # If the clear button isn't present, then there's no edit.
        if self._clearButton.isHidden():
            self._primViewItem.drawModeWidget = None
            self._primViewItem.treeWidget().closePersistentEditor(
                    self._primViewItem, PrimViewColumnIndex.DRAWMODE)

    def _PopupHidden(self):
        # Schedule closing the editor if no edit was made.
        QtCore.QTimer.singleShot(0, self._CloseEditorIfNoEdit)

class DrawModeItemDelegate(QtWidgets.QStyledItemDelegate):
    def __init__(self, makeTimer, parent=None):
        QtWidgets.QStyledItemDelegate.__init__(self, parent=parent)
        self._treeWidget = parent
        self._makeTimer = makeTimer

    # We need to override paint in this delegate as well so that the
    # Draw Mode column will match with the behavior of the other
    # items in the treeview. 
    def paint(self, painter, option, index):
        primViewItem = self._treeWidget.itemFromIndex(index)
        background = _GetBackgroundColor(primViewItem, option)
        if background:
            painter.fillRect(option.rect, background)

        super(DrawModeItemDelegate, self).paint(painter, option, index)

    def createEditor(self, parent, option, index):
        primViewItem = self._treeWidget.itemFromIndex(index)

        if not primViewItem.supportsDrawMode:
            return None

        drawModeWidget = DrawModeWidget(primViewItem, 
            refreshFunc=self._treeWidget.UpdatePrimViewDrawMode,
            makeTimer=self._makeTimer,
            parent=parent)
        # Store a copy of the widget in the primViewItem, for use when 
        # propagating changes to draw mode down a prim hierarchy.
        primViewItem.drawModeWidget = drawModeWidget
        return drawModeWidget

class SelectedAncestorItemDelegate(QtWidgets.QStyledItemDelegate):
    def __init__(self, parent=None):
        QtWidgets.QStyledItemDelegate.__init__(self, parent=parent)
        self._treeWidget = parent

    # In order to highlight the ancestors of selected prims, we require
    # a new delegate to be created to override the paint function.
    # Because the stylesheet will override styling when items are hovered
    # over, the hovering styling logic is moved to this delegate, and 
    # primTreeWidget is excluded from the selectors for hovering
    # in the stylesheet. 
    def paint(self, painter, option, index):
        primViewItem = self._treeWidget.itemFromIndex(index)

        originalPosition = option.rect.left()
        offsetPosition = self._treeWidget.header().offset()

        # In order to fill in the entire cell for Prim Name, we must update the 
        # dimensions of the rectangle painted. If the column is for Prim Name,
        # we set the left side of the rectangle to be equal to the offset of the
        # tree widget's header. 
        background = _GetBackgroundColor(primViewItem, option)

        if primViewItem.ancestorOfSelected:
            if index.column() == PrimViewColumnIndex.NAME:
                option.rect.setLeft(offsetPosition)

        if background:
            painter.fillRect(option.rect, background)

        # resetting the dimensions of the rectangle so that we paint the correct
        # content in the cells on top of the colors we previously painted.
        option.rect.setLeft(originalPosition)
        super(SelectedAncestorItemDelegate, self).paint(painter, option, index)


class PrimItemSelectionModel(QtCore.QItemSelectionModel):
    """Standard QItemSelectionModel does not allow one to have full-item
    selection while exlcuding some columns in the view from activating
    selection.  Since that's exactly the behavior we want, we derive our
    own class that we can force to ignore selection requests except when we
    really want it to."""
    
    def __init__(self, model):
        super(PrimItemSelectionModel, self).__init__(model)
        self._processSelections = True

    @property
    def processSelections(self):
        """If True, calls to clear(), reset(), and select() will function
        as normal.  If False, such calls will be ignored."""
        return self._processSelections

    @processSelections.setter
    def processSelections(self, doProcess):
        self._processSelections = doProcess

    def clear(self):
        if self.processSelections:
            super(PrimItemSelectionModel, self).clear()

    def reset(self):
        if self.processSelections:
            super(PrimItemSelectionModel, self).reset()

    def select(self, indexOrSelection, command):
        if self.processSelections:
            super(PrimItemSelectionModel, self).select(indexOrSelection, command)


class SelectionEnabler(object):
    def __init__(self, selectionModel):
        self._selectionModel = selectionModel
        self._selectionWasEnabled = False

    def __enter__(self):
        self._selectionWasEnabled = self._selectionModel.processSelections
        self._selectionModel.processSelections = True
        return self

    def __exit__(self, *args):
        self._selectionModel.processSelections = self._selectionWasEnabled

# This class extends QTreeWidget and is used to draw the prim tree view in 
# usdview. 
# More of the prim browser specific behavior needs to be migrated from 
# appController into this class.
class PrimTreeWidget(QtWidgets.QTreeWidget):
    def __init__(self, parent):
        super(PrimTreeWidget, self).__init__(parent=parent)
        self._appController = None
        self._selectionModel = PrimItemSelectionModel(self.model())
        self.setSelectionModel(self._selectionModel)
        # The list of ancestors of currently selected items
        self._ancestorsOfSelected = []

    def InitControllers(self, appController):
        self._appController = appController
        selectedAncestorItemDelegate = SelectedAncestorItemDelegate(parent=self)
        self.setItemDelegate(selectedAncestorItemDelegate)
        drawModeItemDelegate = DrawModeItemDelegate(appController._makeTimer,
                                                    parent=self)
        self.setItemDelegateForColumn(PrimViewColumnIndex.DRAWMODE, 
                                      drawModeItemDelegate)

    def ShowDrawModeWidgetForItem(self, primViewItem):
        self.openPersistentEditor(primViewItem, PrimViewColumnIndex.DRAWMODE)

    def UpdatePrimViewDrawMode(self, rootItem=None):
        """Updates browser's "Draw Mode" columns."""
        with self._appController._makeTimer("update draw mode column"):
            self.setUpdatesEnabled(False)
            # Update draw-model for the entire prim tree if the given
            # rootItem is None.
            if rootItem is None:
                rootItem = self.invisibleRootItem().child(0)
            if rootItem.childCount() == 0:
                self._appController._populateChildren(rootItem)
            rootsToProcess = [rootItem.child(i) for i in 
                    range(rootItem.childCount())]
            for item in rootsToProcess:
                PrimViewItem.propagateDrawMode(item, self)
            self.setUpdatesEnabled(True)

    def ColumnPressCausesSelection(self, col):
        """If this method returns True for column `col`, then we want a
        click in that column to cause the item to be selected."""
        return col != PrimViewColumnIndex.VIS and \
            col != PrimViewColumnIndex.GUIDES and \
            col != PrimViewColumnIndex.DRAWMODE

    def ExpandItemRecursively(self, item):
        if item.parent() is not None:
            item = item.parent()
            while item.parent():
                if not item.isExpanded():
                    self.expandItem(item)
                item = item.parent()

    def FrameSelection(self):
        if (self._appController):
            selectedItems = [
                self._appController._getItemAtPath(prim.GetPath())
                for prim in self._appController._dataModel.selection.getPrims()]

            for item in selectedItems:
                self.ExpandItemRecursively(item)

            self.scrollToItem(selectedItems[0])

    # We set selectability based on the column we mousePress'd in, and then
    # restore to true when leaving the widget, so that when we're not
    # interacting with the browser, anyone can modify selection through the
    # regular API's, which is important for the appController and other
    # clients.  If we retore it *before* leaving the widget, some internal
    # QTreeWidget mechanism (an event filter?) _occasionally_ selects the item
    # after a mouse release!
    def mousePressEvent(self, ev):
        item = self.itemAt(QtCore.QPoint(ev.x(), ev.y()))
        if item:
            col = self.columnAt(ev.x())
            self._selectionModel.processSelections = self.ColumnPressCausesSelection(col)
        super(PrimTreeWidget, self).mousePressEvent(ev)

    def leaveEvent(self, ev):
        super(PrimTreeWidget, self).leaveEvent(ev)
        self._selectionModel.processSelections = True

    # We override these selection and interaction-related API, and provide a
    # batch wrapper for QTreeWidgetItem.setSelected in case, in the future,
    # we have user plugins firing while we are still interacting with this
    # widget, and they manipulate selection.
    def clearSelection(self):
        self._resetAncestorsOfSelected()
        with SelectionEnabler(self._selectionModel):
            super(PrimTreeWidget, self).clearSelection()

    def reset(self):
        self._resetAncestorsOfSelected()
        with SelectionEnabler(self._selectionModel):
            super(PrimTreeWidget, self).reset()

    def selectAll(self):
        self._resetAncestorsOfSelected()
        with SelectionEnabler(self._selectionModel):
            super(PrimTreeWidget, self).selectAll()

    def keyPressEvent(self, ev):
        with SelectionEnabler(self._selectionModel):
            # Because setCurrentItem autoexpands the primview so that
            # the current item is visible, we must set the current item
            # only when we know it'll be needed for arrow navigation.
            # We call this here before the selection data model clears
            # the selection. 

            if ev.key() == QtCore.Qt.Key.Key_Down \
            or ev.key() == QtCore.Qt.Key.Key_Up \
            or ev.key() == QtCore.Qt.Key.Key_Right \
            or ev.key() == QtCore.Qt.Key.Key_Left:
                currentPrim = self._appController._dataModel.selection.getFocusPrim()
                currentItem = self._appController._getItemAtPath(currentPrim.GetPath())
                self.setCurrentItem(currentItem, 0, QtCore.QItemSelectionModel.SelectionFlag.NoUpdate)

            super(PrimTreeWidget, self).keyPressEvent(ev)

            # Note that the default behavior in the super class's keyPressEvent
            # function is to set the current item to the first item found that 
            # alphabetically matches the key entered via the keyboardSearch 
            # method. Since we want to override this behavior for the F hotkey
            # to frame the selection, we also need to override keyboardSearch
            # method to disable it. 
            if ev.key() == KeyboardShortcuts.FramingKey:
                self.FrameSelection()

    def keyReleaseEvent(self, ev):
        with SelectionEnabler(self._selectionModel):
            super(PrimTreeWidget, self).keyReleaseEvent(ev)

    def keyboardSearch(self, s):
        """Disable keyboardSearch in the prim view widget. This would interfere
        with F key used for framing the current selection by trying to change
        the selection to the first item starting with F. Since we're disabling
        this for F, it makes sense to disable it for all letters"""
        pass

    def updateSelection(self, added, removed):
        """Mutate the widget's selected items, selecting items in `added`
        and deselecting items in `removed`.  Prefer this method for client
        use over calling setSelected directly on PrimViewItems."""
        with SelectionEnabler(self._selectionModel):
            for item in added:
                item.setSelected(True)
            for item in removed:
                item.setSelected(False)
        self._refreshAncestorsOfSelected()
        # This is a big hammer... if we instead built up a list of the
        # ModelIndices of all the changed ancestors, we could instead make
        # our selectionModel emit selectionChanged for just those items,
        # instead.  Does not currently seem to be impacting interactivity.
        QtWidgets.QWidget.update(self)

    def _resetAncestorsOfSelected(self):
        for item in self._ancestorsOfSelected:
            item.ancestorOfSelected = False
        self._ancestorsOfSelected = []

    # Refresh the list of ancestors of selected prims
    def _refreshAncestorsOfSelected(self):
        selectedItems = [
            self._appController._getItemAtPath(prim.GetPath())
            for prim in self._appController._dataModel.selection.getPrims()]

        self._resetAncestorsOfSelected()

        # Add all ancestor of the prim associated with the item
        # to _ancestorsOfSelected, and mark them as selected ancestors
        for item in selectedItems:
            while item.parent():
                item.parent().ancestorOfSelected = True
                self._ancestorsOfSelected.append(item.parent())
                item = item.parent()



#
# Copyright 2022 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#

import mmap
import os
import shutil
import sys
import tempfile
import zipfile

def _Print(msg):
    print(msg)

def _Err(msg):
    sys.stderr.write(msg + '\n')

def _AllowedUsdzExtensions():
    return [".usdz"]

def _AllowedUsdExtensions():
    return [".usd", ".usda", ".usdc"]


def ExtractUsdzPackage(usdzFile, extractDir, verbose=False, force=False):
    """Extract a .usdz package usdzFile to extractDir.

    Entries are written at their archive-relative paths verbatim.  Nested .usdz
    files are left as-is inside the extract dir; callers that need to walk their
    contents should use UsdzScanIterator or UsdzUpdateIterator, both of which
    handle nesting natively.

    Returns True on success, False on any error.

    """
    if not usdzFile.endswith('.usdz'):
        if verbose:
            _Print("'%s' does not have .usdz extension" % usdzFile)
        return False

    if not os.path.exists(usdzFile):
        if verbose:
            _Print("usdz file '%s' does not exist." % usdzFile)
        return False

    if not extractDir:
        if verbose:
            _Print("No extract dir specified")
        return False

    absExtractDir = os.path.abspath(extractDir)

    if os.path.exists(absExtractDir) and not os.path.isdir(absExtractDir):
        if verbose:
            _Print("Extract path '%s' exists and is not a directory." %
                   extractDir)
        return False

    if force and os.path.isdir(absExtractDir):
        shutil.rmtree(absExtractDir)

    if os.path.isdir(absExtractDir):
        if verbose:
            _Print("Extract Dir: '%s' already exists." % extractDir)
        return False

    parent = os.path.dirname(absExtractDir) or '.'
    tmpExtractPath = tempfile.mkdtemp(dir=parent)
    try:
        with zipfile.ZipFile(usdzFile) as usdzArchive:
            if verbose:
                _Print("Extracting usdz file '%s' to '%s'" %
                       (usdzFile, absExtractDir))
            usdzArchive.extractall(tmpExtractPath)
        os.rename(tmpExtractPath, absExtractDir)
        return True
    except Exception as e:
        _Err("Failed to extract usdz '%s': %s" % (usdzFile, e))
        shutil.rmtree(tmpExtractPath, ignore_errors=True)
        return False

class UsdzScanIterator:
    """Read-only iterator over the entries of a .usdz package.

    Within context, yields (arcname, memoryview) pairs for each file entry.
    Nested .usdz packages are recursed into transparently; arcnames inside
    nested packages use '/' to join boundaries (e.g. 'inner.usdz/foo.usdc').

    The memoryview is a zero-copy slice over an mmap of the outermost package
    file.  It is invalid after __exit__; to retain bytes past the context, copy
    explicitly via bytes(view) or view.tobytes().

    The zip directory itself is parsed via Sdf.ZipFile (we read offsets from it
    but slice bytes out of our own mmap, so GetFile's bytes-copying path is
    never taken).  Nested .usdz packages are spilled to a temp file just long
    enough to call Sdf.ZipFile.Open on them; the bytes we yield are still slices
    of the outer mmap.

    """

    def __init__(self, usdzPath, *, recurse=True, verbose=False):
        self._path = usdzPath
        self._recurse = recurse
        self._verbose = verbose
        self._fd = None
        self._mmap = None
        self._views = []
        self._tmpFiles = []

    def __enter__(self):
        self._fd = open(self._path, 'rb')
        try:
            size = os.fstat(self._fd.fileno()).st_size
            if size == 0:
                raise ValueError("'%s' is empty" % self._path)
            self._mmap = mmap.mmap(
                self._fd.fileno(), 0, access=mmap.ACCESS_READ)
        except Exception:
            self._fd.close()
            self._fd = None
            raise
        return self

    def __exit__(self, excType, excVal, excTB):
        for v in self._views:
            try:
                v.release()
            except Exception:
                pass
        self._views = []
        for tmp in self._tmpFiles:
            try:
                os.unlink(tmp)
            except OSError:
                pass
        self._tmpFiles = []
        if self._mmap is not None:
            self._mmap.close()
            self._mmap = None
        if self._fd is not None:
            self._fd.close()
            self._fd = None

    def __iter__(self):
        if self._mmap is None:
            raise RuntimeError("UsdzScanIterator used outside of its context")
        from pxr import Sdf
        zipFile = Sdf.ZipFile.Open(self._path)
        if zipFile is None:
            raise ValueError("Sdf.ZipFile.Open failed for '%s'" % self._path)
        yield from self._Walk(zipFile, 0, "")

    def _Walk(self, zipFile, baseOffset, prefix):
        for name in zipFile.GetFileNames():
            if name.endswith('/'):
                continue
            info = zipFile.GetFileInfo(name)
            if info is None:
                continue
            if info.encrypted:
                if self._verbose:
                    _Print("Skipping encrypted entry: %s" % name)
                continue
            if info.compressionMethod != 0:
                if self._verbose:
                    _Print("Skipping non-stored entry (method %d): %s"
                           % (info.compressionMethod, name))
                continue
            dataAbs = baseOffset + info.dataOffset
            full = prefix + name
            if self._recurse and name.lower().endswith('.usdz'):
                if self._verbose:
                    _Print("Recursing into nested usdz: %s" % full)
                yield from self._WalkNested(dataAbs, info.size, full)
            else:
                view = memoryview(self._mmap)[dataAbs:dataAbs + info.size]
                self._views.append(view)
                yield (full, view)

    def _WalkNested(self, dataAbs, size, arcname):
        from pxr import Sdf
        tmpFd, tmpPath = tempfile.mkstemp(
            suffix='.usdz', prefix='.usdzScan.%d.' % os.getpid())
        try:
            os.write(tmpFd, bytes(self._mmap[dataAbs:dataAbs + size]))
        finally:
            os.close(tmpFd)
        self._tmpFiles.append(tmpPath)
        nested = Sdf.ZipFile.Open(tmpPath)
        if nested is None:
            if self._verbose:
                _Print("Could not open nested usdz: %s" % arcname)
            return
        yield from self._Walk(nested, dataAbs, arcname + '/')


class UsdzUpdateIterator:
    """Extract a .usdz package on enter; on clean exit, atomically rewrite the
    package (optionally to a different output path) from the (possibly modified)
    extracted contents.

    Atomic rewrite: pack to a sibling temp file (see naming below),
    optionally chmod/chown to match the original, then os.replace() into
    place.  The original is left untouched on any failure along the way.

    If the with-block raises, the rewrite is skipped entirely.

    Temp-file naming:
        extract dir:    .usdzExtract[.<tag>].<pid>.<random>/
        rewrite temp:   <output>[.<tag>].<pid>.tmp.usdz
    The optional 'tag' argument lets callers (e.g. a campaign script)
    inject a distinctive marker so abandoned temps from a crash can be
    identified unambiguously by name.  Tag should be a simple identifier
    (letters/digits/dashes/underscores); callers are responsible for not
    passing path separators or bracket characters.

    Platform notes:
    - preserveAttrs is POSIX-only and rejected on Windows; os.chown is not
      available and os.chmod only honors the read-only bit there, so there's no
      meaningful permission/ownership to copy.
    - On Windows, os.replace fails if another process holds the destination open
      (sharing violation).  Callers that need to tolerate that should retry
      around the with-block.

    """

    def __init__(self, usdzPath, *, outputPath=None, preserveAttrs=False,
                 parentDir=None, tag=None, verbose=False):
        if preserveAttrs and os.name == 'nt':
            raise ValueError("preserveAttrs=True is not supported on Windows")
        self._inputPath = os.path.abspath(usdzPath)
        self._outputPath = (os.path.abspath(outputPath) if outputPath
                            else self._inputPath)
        self._preserveAttrs = preserveAttrs
        self._parentDir = parentDir
        self._tag = tag
        self._verbose = verbose
        self._extractDir = None

    def _TaggedInfix(self):
        """'.<tag>.' if tagged, else '.'.  Used in temp-name construction."""
        return f".{self._tag}." if self._tag else "."

    def __enter__(self):
        # Both branches use mkdtemp so two same-stem nested packages
        # (e.g. dirA/nested.usdz and dirB/nested.usdz inside the same outer) get
        # distinct extract dirs.  Previously the parentDir branch derived the
        # extract dir name from the input file's stem; that was safe only under
        # strictly sequential iteration.
        prefix = '.usdzExtract%s%d.' % (self._TaggedInfix(), os.getpid())
        if self._parentDir is not None:
            parent = self._parentDir
        else:
            parent = os.path.dirname(self._outputPath) or '.'
        self._extractDir = tempfile.mkdtemp(prefix=prefix, dir=parent)

        if not ExtractUsdzPackage(self._inputPath, self._extractDir,
                                  verbose=self._verbose, force=True):
            raise RuntimeError(
                "Failed to extract usdz package '%s' to '%s'"
                % (self._inputPath, self._extractDir))
        return self

    def __exit__(self, excType, excVal, excTB):
        try:
            if excType is not None:
                if self._verbose:
                    _Print("UsdzUpdateIterator: skipping rewrite due "
                           "to exception (%s)" % excType.__name__)
                return

            if not self._extractDir or not os.path.isdir(self._extractDir):
                return

            origStat = None
            if self._preserveAttrs and os.path.exists(self._inputPath):
                origStat = os.stat(self._inputPath)

            tmpPath = "%s%s%d.tmp.usdz" % (
                self._outputPath, self._TaggedInfix(), os.getpid())

            self._WritePackage(tmpPath)

            if origStat is not None:
                try:
                    os.chmod(tmpPath, origStat.st_mode)
                    os.chown(tmpPath, origStat.st_uid, origStat.st_gid)
                except OSError:
                    try:
                        os.unlink(tmpPath)
                    except OSError:
                        pass
                    raise

            os.replace(tmpPath, self._outputPath)
        finally:
            if self._extractDir and os.path.isdir(self._extractDir):
                shutil.rmtree(self._extractDir, ignore_errors=True)

    def _WritePackage(self, outPath):
        from pxr import Sdf, Tf
        with Sdf.ZipFileWriter.CreateNew(outPath) as usdzWriter:
            for absPath, arcname in self.Entries():
                if self._verbose:
                    _Print('.. adding: %s' % arcname)
                try:
                    usdzWriter.AddFile(absPath, arcname)
                except Tf.ErrorException:
                    _Err("Failed to add file '%s' (as '%s') to package. "
                         "Discarding." % (absPath, arcname))
                    raise

    def Entries(self):
        """Yield (absPath, arcname) for every regular file under the extract
        dir.  Flat: nested .usdz packages appear as a single file entry, not
        their contents.  Arcnames use '/' separators."""
        if not self._extractDir or not os.path.isdir(self._extractDir):
            return
        for root, _dirs, files in os.walk(self._extractDir):
            for f in files:
                absPath = os.path.join(root, f)
                rel = os.path.relpath(absPath, self._extractDir)
                arcname = rel.replace(os.sep, '/')
                yield absPath, arcname

    def UsdAssets(self):
        """Yield absolute paths to all usd/usda/usdc/usdz assets under the
        extract dir.  Nested .usdz packages are recursed into via a fresh
        nested iterator."""
        allowed = _AllowedUsdzExtensions() + _AllowedUsdExtensions()
        for absPath, arcname in list(self.Entries()):
            ext = os.path.splitext(arcname)[1]
            if ext not in allowed:
                continue
            if ext in _AllowedUsdzExtensions():
                if self._verbose:
                    _Print("Iterating nested usdz asset: %s" % arcname)
                with UsdzUpdateIterator(absPath,
                                        parentDir=self._extractDir,
                                        verbose=self._verbose) as nested:
                    yield from nested.UsdAssets()
            else:
                if self._verbose:
                    _Print("Iterating usd asset: %s" % arcname)
                yield absPath

    def AllAssets(self):
        """Yield absolute paths to every asset under the extract dir, recursing
        transparently into nested .usdz packages."""
        for absPath, arcname in list(self.Entries()):
            ext = os.path.splitext(arcname)[1]
            if ext in _AllowedUsdzExtensions():
                if self._verbose:
                    _Print("Iterating nested usdz asset: %s" % arcname)
                with UsdzUpdateIterator(absPath,
                                        parentDir=self._extractDir,
                                        verbose=self._verbose) as nested:
                    yield from nested.AllAssets()
            else:
                if self._verbose:
                    _Print("Iterating usd asset: %s" % arcname)
                yield absPath


class UsdzAssetIterator(UsdzUpdateIterator):
    """Deprecated: use UsdzUpdateIterator for new code.

    Thin compatibility shim preserving the legacy positional signature
    UsdzAssetIterator(usdzFile, verbose, parentDir=None).  Behaves like
    UsdzUpdateIterator with outputPath=usdzFile and preserveAttrs=False.
    """
    def __init__(self, usdzFile, verbose, parentDir=None):
        super().__init__(usdzFile,
                         outputPath=None,
                         preserveAttrs=False,
                         parentDir=parentDir,
                         verbose=verbose)
